@@ -20,6 +20,7 @@ from glob import glob
 from pathlib import Path
 from typing import (
     IO,
+    TYPE_CHECKING,
     Any,
     Callable,
     Collection,
@@ -31,7 +32,6 @@ from typing import (
     Tuple,
     Union,
     cast,
-    TYPE_CHECKING,
 )
 from warnings import warn
 
@@ -43,15 +43,15 @@ from pandas.core.base import DataError
 from qslib.plate_setup import PlateSetup
 from qslib.scpi_commands import AccessLevel, SCPICommand
 
-from .version import __version__
+from ._analysis_protocol_text import _ANALYSIS_PROTOCOL_TEXT
+from ._util import _nowuuid, _pp_seqsliceint, _set_or_create
 from .base import RunStatus
 from .data import FilterDataReading, FilterSet, df_from_readings
 from .machine import Machine
-from .normalization import Normalizer, NormRaw
-from .rawquant_compat import _fdc_to_rawdata
+from .processors import NormRaw, Processor
 from .protocol import Protocol, Stage, Step
-from ._util import _set_or_create, _nowuuid, _pp_seqsliceint
-from ._analysis_protocol_text import _ANALYSIS_PROTOCOL_TEXT
+from .rawquant_compat import _fdc_to_rawdata
+from .version import __version__
 
 if TYPE_CHECKING:  # pragma: no cover
     import matplotlib.pyplot as plt
@@ -294,9 +294,23 @@ class Experiment:
 
     @property
     def all_filters(self) -> Collection[FilterSet]:
-        """All filters used at some point in the experiment's protocol."""
+        """All filters used at some point in the experiment.
+
+        If the experiment has data, this is based on the existing data.  Otherwise, it is based
+        on the experiment protocol.
+        """
+        if self._welldata is not None:
+            return [
+                FilterSet.fromstring(f)
+                for f in self.welldata.index.get_level_values(0).unique()
+            ]
         return self.protocol.all_filters
 
+    @property
+    def filter_strings(self) -> list[str]:
+        """All filters, as x?-m? strings, used at some point in the experiment."""
+        return [str(f) for f in self.all_filters]
+
     @property
     def welldata(self) -> pd.DataFrame:
         """
@@ -455,6 +469,9 @@ table, th, td {{
     ) -> Machine:
         if isinstance(machine, Machine):
             self.machine = machine
+            self.machine.max_access_level = max(
+                self.machine.max_access_level, needed_level
+            )
             return machine
         elif isinstance(machine, str):
             self.machine = Machine(
@@ -464,6 +481,7 @@ table, th, td {{
             )
             return self.machine
         elif hasattr(self, "machine") and (c := self.machine):
+            c.max_access_level = max(c.max_access_level, needed_level)
             return c
         else:
             raise ValueError(
@@ -703,6 +721,48 @@ table, th, td {{
             # The message log is tricky. Ideally we'd use rsync or wc+tail. TODO
             self._update_from_files()
 
+    def change_protocol_from_now(
+        self, new_stages: Sequence[Stage], machine: MachineReference | None = None
+    ) -> None:
+        """
+        For a running experiment, change the remaining stages to be the provided :param:`new_stages` list.
+        This is a convenience function that:
+
+        1. Gets the currently-running stage and cycle.
+        2. Sets the repeat number of the current stage to its current cycle, thus ending it after the end of
+           the current cycle.
+        3. Changes the remainder of the stages to be those in the :param:`new_stages` list.
+
+        Because this does not impact any current or past stages, there is less of a need to ensure that the
+        stages provided are compatible with the old protocol.  The only check done is to ensure that, if
+        the provided stages have any collection commands using default filters, the old protocol has specified
+        default filters.  This function does not allow the default filters to be changed: if you want to use
+        filters other than the defaults, or defaults were not provided in the old protocol, then either specify
+        filters explicitly (recommended) for new stages you'd like to be different, or use
+        :any:`Experiment.change_protocol` directly.
+        """
+        machine = self._ensure_machine(machine, needed_level=AccessLevel.Controller)
+        with machine.ensured_connection(AccessLevel.Controller):
+            self._ensure_running(machine)
+
+            proto = machine.get_running_protocol()
+
+            runstatus = machine.run_status()
+
+            proto.stages[-1].repeat = runstatus.cycle
+            proto.stages += new_stages
+
+            machine.define_protocol(proto)
+            self.protocol = proto
+
+            self._update_tcprotocol_xml()
+
+            # Push new tcprotocol.xml
+            machine.write_file(
+                "${LogFolder}/tcprotocol.xml",
+                open(self._sdspath("tcprotocol.xml"), "rb").read(),
+            )
+
     def change_protocol(
         self,
         new_protocol: Protocol,
@@ -832,7 +892,8 @@ table, th, td {{
 
     @sample_wells.setter
     def sample_wells(self, new_sample_wells: dict[str, list[str]]) -> None:
-        self.plate_setup.sample_wells = new_sample_wells
+        raise NotImplementedError
+        # self.plate_setup.sample_wells = new_sample_wells
 
     def __init__(
         self,
@@ -1332,7 +1393,7 @@ table, th, td {{
         self.runstate = "INIT"
 
         m: Optional[re.Match[str]]
-        stages = []
+        stages: list[dict[str, int | str | datetime]] = []
 
         for m in ms:
             ts = datetime.fromtimestamp(float(m["ts"]))
@@ -1347,28 +1408,28 @@ table, th, td {{
                 elif m["ext"] == "POSTRun":
                     self.activeendtime = ts
                 try:
-                    stages.append([int(m["msg"]), ts])
+                    stages.append({"stage": int(m["ext"]), "start_time": ts})
                 except ValueError:
-                    stages.append([m["ext"], ts])
+                    stages.append({"stage": m["ext"], "start_time": ts})
                 if len(stages) > 1:
-                    stages[-2].append(ts)
+                    stages[-2]["end_time"] = ts
             elif m["msg"] == "Ended":
                 self.runendtime = ts
                 self.runstate = "COMPLETE"
                 if len(stages) > 1:
-                    stages[-1].append(ts)
+                    stages[-1].setdefault("end_time", ts)
             elif m["msg"] == "Aborted":
                 self.runstate = "ABORTED"
                 self.activeendtime = ts
                 self.runendtime = ts
                 if len(stages) > 1:
-                    stages[-1].append(ts)
+                    stages[-1].setdefault("end_time", ts)
             elif m["msg"] == "Stopped":
                 self.runstate = "STOPPED"
                 self.activeendtime = ts
                 self.runendtime = ts
                 if len(stages) > 1:
-                    stages[-1].append(ts)
+                    stages[-1].setdefault("end_time", ts)
 
         self.stages = pd.DataFrame(stages, columns=["stage", "start_time", "end_time"])
 
@@ -1503,7 +1564,8 @@ table, th, td {{
         anneal_stages: int | Sequence[int] | None = None,
         melt_stages: int | Sequence[int] | None = None,
         between_stages: int | Sequence[int] | None = None,
-        normalization: Normalizer = NormRaw(),
+        process: Sequence[Processor] | Processor | None = None,
+        normalization: Processor | None = None,
         ax: "plt.Axes" | None = None,
         marker: str | None = None,
         legend: bool | Literal["inset", "right"] = True,
@@ -1582,6 +1644,15 @@ table, th, td {{
 
         import matplotlib.pyplot as plt
 
+        if process is None:
+            process = [normalization or NormRaw()]
+        elif normalization:
+            raise ValueError(
+                "Can't specify both process and normalization (include normalization in process list)."
+            )
+        if isinstance(process, Processor):
+            process = [process]
+
         if filters is None:
             filters = self.all_filters
 
@@ -1621,13 +1692,17 @@ table, th, td {{
                 ).add_subplot(),
             )
 
-        data = normalization.normalize_scoped(self.welldata, "all")
+        data = self.welldata
+
+        for processor in process:
+            data = processor.process_scoped(data, "all")
 
         all_wells = self.plate_setup.get_wells(samples)
 
         reduceddata = data.loc[[f.lowerform for f in filters], all_wells]
 
-        reduceddata = normalization.normalize_scoped(reduceddata, "limited")
+        for processor in process:
+            reduceddata = processor.process_scoped(reduceddata, "limited")
 
         for filter in filters:
             filterdat: pd.DataFrame = reduceddata.loc[filter.lowerform, :]  # type: ignore
@@ -1648,7 +1723,14 @@ table, th, td {{
                 for well in wells:
                     color = next(ax._get_lines.prop_cycler)["color"]
 
-                    label = _gen_label(sample, well, filter, samples, wells, filters)
+                    label = _gen_label(
+                        self.plate_setup.get_descriptive_string(sample),
+                        well,
+                        filter,
+                        samples,
+                        wells,
+                        filters,
+                    )
 
                     anneallines.append(
                         ax.plot(
@@ -1686,7 +1768,10 @@ table, th, td {{
 
         ax.set_xlabel("temperature (°C)")
 
-        ax.set_ylabel(normalization.ylabel)
+        ylabel: str | None = None
+        for processor in process:
+            ylabel = processor.ylabel(ylabel)
+        ax.set_ylabel(ylabel or "fluorescence")
 
         if legend is True:
             if len(anneallines) < 6:
@@ -1732,7 +1817,8 @@ table, th, td {{
         samples: str | Sequence[str] | None = None,
         filters: str | FilterSet | Collection[str | FilterSet] | None = None,
         stages: slice | int | Sequence[int] = slice(None),
-        normalization: Normalizer = NormRaw(),
+        process: Sequence[Processor] | Processor | None = None,
+        normalization: Processor | None = None,
         ax: "plt.Axes" | "Sequence[plt.Axes]" | None = None,
         legend: bool | Literal["inset", "right"] = True,
         temperatures: Literal[False, "axes", "inset", "twin"] = "axes",
@@ -1832,6 +1918,15 @@ table, th, td {{
 
         import matplotlib.pyplot as plt
 
+        if process is None:
+            process = [normalization or NormRaw()]
+        elif normalization:
+            raise ValueError(
+                "Can't specify both process and normalization (include normalization in process list)."
+            )
+        if isinstance(process, Processor):
+            process = [process]
+
         if filters is None:
             filters = self.all_filters
 
@@ -1865,13 +1960,17 @@ table, th, td {{
 
         ax = cast(Sequence[plt.Axes], ax)
 
-        data = normalization.normalize_scoped(self.welldata, "all")
+        data = self.welldata
+
+        for processor in process:
+            data = processor.process_scoped(data, "all")
 
         all_wells = self.plate_setup.get_wells(samples) + ["time"]
 
         reduceddata = data.loc[[f.lowerform for f in filters], all_wells]
 
-        reduceddata = normalization.normalize_scoped(reduceddata, "limited")
+        for processor in process:
+            reduceddata = processor.process_scoped(reduceddata, "limited")
 
         lines = []
         for filter in filters:
@@ -1883,7 +1982,14 @@ table, th, td {{
                 for well in wells:
                     color = next(ax[0]._get_lines.prop_cycler)["color"]
 
-                    label = _gen_label(sample, well, filter, samples, wells, filters)
+                    label = _gen_label(
+                        self.plate_setup.get_descriptive_string(sample),
+                        well,
+                        filter,
+                        samples,
+                        wells,
+                        filters,
+                    )
 
                     lines.append(
                         ax[0].plot(
@@ -1898,7 +2004,10 @@ table, th, td {{
 
         ax[-1].set_xlabel("time (hours)")
 
-        ax[0].set_ylabel(normalization.ylabel)
+        ylabel: str | None = None
+        for processor in process:
+            ylabel = processor.ylabel(ylabel)
+        ax[0].set_ylabel(ylabel or "fluorescence")
 
         if legend is True:
             if len(lines) < 6:
@@ -2106,10 +2215,17 @@ def _normalize_filters(
     return [FilterSet.fromstring(filter) for filter in filters]
 
 
-def _gen_label(sample, well, filter, samples, wells, filters) -> str:
+def _gen_label(
+    sample: str,
+    well: str,
+    filter: FilterSet,
+    samples: Sequence[str],
+    wells: Sequence[str],
+    filters: Sequence[FilterSet],
+) -> str:
     label = ""
     if len(samples) > 1:
-        label = str(sample)
+        label = sample
     if len(wells) > 1:
         if len(label) > 0:
             label += f" ({well})"
@@ -2124,8 +2240,14 @@ def _gen_label(sample, well, filter, samples, wells, filters) -> str:
     return label
 
 
-def _gen_axtitle(expname, stages, samples, wells, filters) -> str:
-    elems = []
+def _gen_axtitle(
+    expname: str,
+    stages: Sequence[int] | slice | int,
+    samples: Sequence[str],
+    wells: Sequence[str],
+    filters: Sequence[FilterSet],
+) -> str:
+    elems: list[str] = []
     if len(samples) == 1:
         elems += samples
     if len(filters) == 1:
