@@ -13,16 +13,16 @@ from asyncio.futures import Future
 from contextlib import contextmanager
 from dataclasses import dataclass
 from functools import wraps
-from typing import IO, Any, Generator, Literal, overload, TYPE_CHECKING, cast
+from typing import IO, TYPE_CHECKING, Any, Generator, Literal, cast, overload
 
 import nest_asyncio
 
 from qslib.qs_is_protocol import CommandError
 from qslib.scpi_commands import AccessLevel, SCPICommand
 
-from .qsconnection_async import QSConnectionAsync
-from .protocol import Protocol
 from ._util import _unwrap_tags
+from .protocol import Protocol
+from .qsconnection_async import QSConnectionAsync
 
 nest_asyncio.apply()
 
@@ -31,14 +31,15 @@ from .base import MachineStatus, RunStatus
 log = logging.getLogger(__name__)
 
 if TYPE_CHECKING:  # pragma: no cover
-    from .experiment import Experiment
     import matplotlib.pyplot as plt
 
+    from .experiment import Experiment
 
-def _ensure_connection(level: AccessLevel = AccessLevel.Observer):
+
+def _ensure_connection(level: AccessLevel = AccessLevel.Observer) -> Any:
     def wrap(func):
         @wraps(func)
-        def wrapped(m: Machine, *args, **kwargs):
+        def wrapped(m: Machine, *args: Any, **kwargs: Any) -> Any:
             if m.automatic:
                 with m.ensured_connection(level):
                     return func(m, *args, **kwargs)
@@ -124,7 +125,7 @@ class Machine:
         return self._max_access_level
 
     @max_access_level.setter
-    def max_access_level(self, v: AccessLevel | str):
+    def max_access_level(self, v: AccessLevel | str) -> None:
         if not isinstance(v, AccessLevel):
             self._max_access_level = AccessLevel(v)
         else:
@@ -173,7 +174,11 @@ class Machine:
             return self.connection.connected
 
     def __enter__(self) -> Machine:
-        self.connect()
+        try:
+            self.connect()
+        except Exception as e:
+            self.disconnect()
+            raise e
         return self
 
     @_ensure_connection(AccessLevel.Guest)
@@ -637,7 +642,7 @@ class Machine:
                 raise ValueError(f"Unexpected power status: {s}")
 
     @power.setter
-    def power(self, value: Literal["on", "off", True, False]):
+    def power(self, value: Literal["on", "off", True, False]) -> None:
         with self.ensured_connection(AccessLevel.Controller):
             if value is True:
                 value = "on"
