@@ -6,18 +6,20 @@ from __future__ import annotations
 
 import logging
 import math
-from os import popen
 import textwrap
+import warnings
 import xml.etree.ElementTree as ET
 from abc import ABC, abstractmethod
 from copy import deepcopy
 from dataclasses import dataclass, field
 from itertools import zip_longest
 from typing import (
+    TYPE_CHECKING,
     Any,
     Callable,
     ClassVar,
     Collection,
+    Generic,
     Iterable,
     List,
     Optional,
@@ -27,7 +29,7 @@ from typing import (
     TypeVar,
     Union,
     cast,
-    TYPE_CHECKING,
+    overload,
 )
 
 import attr
@@ -38,12 +40,10 @@ import pint
 from qslib import scpi_commands
 from qslib.data import FilterSet
 
-from .version import __version__
+from ._util import *
 from .base import RunStatus
 from .scpi_commands import SCPICommand, SCPICommandLike
-from ._util import *
-import warnings
-
+from .version import __version__
 
 if TYPE_CHECKING:  # pragma: no cover
     import matplotlib.pyplot as plt
@@ -183,10 +183,10 @@ T = TypeVar("T")
 
 class ProtoCommand(ABC):
     @abstractmethod
-    def to_scpicommand(self, **kwargs) -> SCPICommand:  # pragma: no cover
+    def to_scpicommand(self, **kwargs: Any) -> SCPICommand:  # pragma: no cover
         ...
 
-    def __init__(self, *args, **kwargs) -> None:  # pragma: no cover
+    def __init__(self, *args: Any, **kwargs: Any) -> None:  # pragma: no cover
         ...
 
     @classmethod
@@ -218,7 +218,7 @@ class Ramp(ProtoCommand):
     )
     _names: ClassVar[Sequence[str]] = ("RAMP",)
 
-    def to_scpicommand(self, **kwargs) -> SCPICommand:
+    def to_scpicommand(self, **kwargs: None) -> SCPICommand:
         opts = {}
 
         if self.increment != _ZEROTEMPDELTA:
@@ -250,7 +250,7 @@ class Exposure(ProtoCommand):
     state: str = "HoldAndCollect"
     _names: ClassVar[Sequence[str]] = ("EXP", "EXPOSURE")
 
-    def to_scpicommand(self, **kwargs) -> SCPICommand:
+    def to_scpicommand(self, **kwargs: None) -> SCPICommand:
         settingstrings = [
             k.hacform + "," + ",".join(str(x) for x in v) for k, v in self.settings
         ]
@@ -285,7 +285,11 @@ class HACFILT(ProtoCommand):
     _default_filters: Sequence[FilterSet] = attr.field(factory=lambda: [])
     _names: ClassVar[Sequence[str]] = ("HoldAndCollectFILTer", "HACFILT")
 
-    def to_scpicommand(self, default_filters=None, **kwargs) -> SCPICommand:
+    def to_scpicommand(
+        self, default_filters: Sequence[FilterSet] | None = None, **kwargs: None
+    ) -> SCPICommand:
+        if default_filters is None:
+            default_filters = []
         if not default_filters and not self.filters:
             raise ValueError("Protocol must have default filters set.")
         if not self.filters:
@@ -320,7 +324,7 @@ class HoldAndCollect(ProtoCommand):
     pcr: bool = False
     _names: ClassVar[Sequence[str]] = ("HoldAndCollect",)
 
-    def to_scpicommand(self, **kwargs) -> SCPICommand:
+    def to_scpicommand(self, **kwargs: None) -> SCPICommand:
         opts = {}
         if self.increment != HoldAndCollect.increment:
             opts["increment"] = self.increment.to("seconds").magnitude
@@ -353,7 +357,7 @@ class Hold(ProtoCommand):
     incrementstep: int = 1
     _names: ClassVar[Sequence[str]] = ("HOLD",)
 
-    def to_scpicommand(self, **kwargs) -> SCPICommand:
+    def to_scpicommand(self, **kwargs: None) -> SCPICommand:
         opts = {}
         if self.increment != Hold.increment:
             opts["increment"] = self.increment.to("seconds").magnitude
@@ -379,12 +383,64 @@ class XMLable(ABC):
         ...
 
 
+G = TypeVar("G")
+
+
+class _NumOrRefIndexer(Generic[G]):
+    _list: list[G]
+
+    def __init__(self, val_list: list[G]):
+        self._list = val_list
+
+    @overload
+    def _translate_key(self, key: int | str) -> int:
+        ...
+
+    @overload
+    def _translate_key(self, key: slice) -> slice:
+        ...
+
+    def _translate_key(self, key: int | str | slice) -> int | slice:
+        if isinstance(key, int):
+            return key - 1
+        elif isinstance(key, str):
+            raise NotImplemented
+        elif isinstance(key, slice):
+            return slice(key.start - 1, key.stop, key.step)
+
+    def __getitem__(self, key: int | str | slice) -> G | list[G]:
+        return self._list[self._translate_key(key)]
+
+    @overload
+    def __setitem__(self, key: int | str, val: G) -> None:
+        ...
+
+    @overload
+    def __setitem__(self, key: slice, val: Sequence[G]) -> None:
+        ...
+
+    def __setitem__(self, key, val):
+        self._list.__setitem__(self._translate_key(key), val)
+
+    def __delitem__(self, key: int | str | slice) -> None:
+        del self._list[self._translate_key(key)]
+
+    def __call__(self, key: int | str | slice) -> G | list[G]:
+        return self[key]
+
+    def append(self, val: G) -> None:
+        return self._list.append(val)
+
+    def __iadd__(self, val: Iterable[G]) -> None:
+        self._list += val
+
+
 class CustomStep(ProtoCommand):
     """A protocol step composed of SCPI/protocol commands."""
 
-    _body: Sequence[ProtoCommand]
+    _body: list[ProtoCommand]
     _identifier: int | str | None = None
-    _repeat: int = 1
+    repeat: int = 1
     _names: ClassVar[Sequence[str]] = ["STEP"]
 
     def __init__(
@@ -393,9 +449,9 @@ class CustomStep(ProtoCommand):
         identifier: int | str | None = None,
         repeat: int = 1,
     ):
-        self._body = body
+        self._body = list(body)
         self._identifier = identifier
-        self._repeat = repeat
+        self.repeat = repeat
 
     def info_str(self, index: None | int = None, repeats: int = 1) -> str:
         if index is not None:
@@ -419,38 +475,30 @@ class CustomStep(ProtoCommand):
         return 0 * UR.seconds
 
     @property
-    def body(self) -> Sequence[ProtoCommand]:
+    def body(self) -> list[ProtoCommand]:
         return self._body
 
     @body.setter
-    def body(self, v: Sequence[ProtoCommand]):
-        self._body = v
-
-    @property
-    def repeat(self) -> int:
-        return self._repeat
-
-    @repeat.setter
-    def repeat(self, v: int):
-        self._repeat = v
+    def body(self, v: Sequence[ProtoCommand]) -> None:
+        self._body = list(v)
 
     @property
     def identifier(self) -> int | str | None:
         return self._identifier
 
     @identifier.setter
-    def identifier(self, v: int | str | None):
+    def identifier(self, v: int | str | None) -> None:
         self._identifier = v
 
     @property
     def collects(self) -> bool:
         return False
 
-    def to_scpicommand(self, *, stepindex: int = -1, **kwargs) -> SCPICommand:
+    def to_scpicommand(self, *, stepindex: int = -1, **kwargs: Any) -> SCPICommand:
         opts = {}
         args: list[int | str | None | Sequence[SCPICommand]] = []
         if self.repeat != 1:
-            opts["repeat"] = self._repeat
+            opts["repeat"] = self.repeat
         if self._identifier:
             args.append(self.identifier)
         else:
@@ -528,12 +576,14 @@ class Step(CustomStep, XMLable):
         on_setattr=attr.setters.convert,
     )
     temp_incrementcycle: int = 2
+    temp_incrementpoint: int = 2
     time_increment: pint.Quantity[int] = attr.field(
         default=UR.Quantity(0, UR.second),
         converter=_wrap_seconds,
         on_setattr=attr.setters.convert,
     )
     time_incrementcycle: int = 2
+    time_incrementpoint: int = 2
     filters: Sequence[FilterSet] = attr.field(
         default=tuple(),
         converter=_filterlist,
@@ -542,10 +592,11 @@ class Step(CustomStep, XMLable):
     pcr: bool = False
     quant: bool = True
     tiff: bool = False
+    repeat: int = 1
     _default_filters: Sequence[FilterSet] = attr.field(default=tuple())
     _classname: ClassVar[str] = "Step"
 
-    def __eq__(self, other: object):
+    def __eq__(self, other: object) -> bool:  # FIXME: add other stuff
         if not isinstance(other, Step):
             return False
         if self.__class__ != other.__class__:
@@ -601,7 +652,9 @@ class Step(CustomStep, XMLable):
         if self.time_increment != 0.0:
             elems.append(f"{_durformat(self.time_increment)}/cycle")
             if self.time_incrementcycle != 2:
-                elems[-1] += f" from cycle {self.time_incrementcycle}"
+                elems[
+                    -1
+                ] += f" from cycle {self.time_incrementcycle}"  # Fixme: ADD OTHER STUFF
         # if self.ramp_rate != 1.6:
         #    elems.append(f"{self.ramp_rate} °C/s ramp")
         s = f"{index}. " + ", ".join(elems)
@@ -627,31 +680,27 @@ class Step(CustomStep, XMLable):
             (self.duration_at_cycle(c) for c in range(1, repeats + 1)), 0 * UR.seconds
         )
 
-    def duration_at_cycle(self, cycle: int) -> pint.Quantity:  # cycle from 1
+    def duration_at_cycle(
+        self, cycle: int
+    ) -> pint.Quantity:  # cycle from 1 # FIXME: add point
         "Duration of the step (excluding ramp) at `cycle` (from 1)"
         inccycles = max(0, cycle + 1 - self.time_incrementcycle)
         return self.time + inccycles * self.time_increment
         # FIXME: is this right?
 
-    def temperatures_at_cycle(self, cycle: int) -> pint.Quantity[np.ndarray]:
+    def temperatures_at_cycle(
+        self, cycle: int
+    ) -> pint.Quantity[np.ndarray]:  # FIXME: add point
         "Temperatures of the step at `cycle` (from 1)"
         inccycles = max(0, cycle + 1 - self.temp_incrementcycle)
         return self.temperature_list + inccycles * self.temp_increment
 
-    @property
-    def repeat(self) -> int:
-        return 1
-
-    @repeat.setter
-    def repeat(self, v: int):
-        raise ValueError
-
     @property
     def identifier(self) -> None:
         return None
 
     @identifier.setter
-    def identifier(self, v: Any):
+    def identifier(self, v: Any) -> None:
         raise ValueError
 
     @property
@@ -667,14 +716,17 @@ class Step(CustomStep, XMLable):
         if self.collects:
             return [
                 Ramp(
-                    self.temperature_list, self.temp_increment, self.temp_incrementcycle
+                    self.temperature_list,
+                    self.temp_increment,
+                    self.temp_incrementcycle,
+                    self.temp_incrementpoint,
                 ),
                 HACFILT(self.filters),
                 HoldAndCollect(
                     self.time,
                     self.time_increment,
                     self.time_incrementcycle,
-                    1,
+                    self.time_incrementpoint,
                     self.tiff,
                     self.quant,
                     self.pcr,
@@ -683,21 +735,27 @@ class Step(CustomStep, XMLable):
         else:
             return [
                 Ramp(
-                    self.temperature_list, self.temp_increment, self.temp_incrementcycle
+                    self.temperature_list,
+                    self.temp_increment,
+                    self.temp_incrementcycle,
+                    self.temp_incrementpoint,
                 ),
                 Hold(
                     self.time,
                     self.time_increment,
                     self.time_incrementcycle,
+                    self.time_incrementpoint,
                 ),
             ]
 
     @body.setter
-    def body(self, v: Any):
+    def body(self, v: Any) -> None:
         raise ValueError
 
     @classmethod
-    def from_xml(cls, e: ET.Element, *, etc=1, ehtc=1, he=False) -> Step:
+    def from_xml(
+        cls, e: ET.Element, *, etc: int = 1, ehtc: int = 1, he: bool = False
+    ) -> Step:
         collect = bool(int(e.findtext("CollectionFlag") or 0))
         ts: pint.Quantity[np.ndarray] = UR.Quantity(
             [float(x.text or math.nan) for x in e.findall("Temperature")], "degC"
@@ -710,7 +768,7 @@ class Step(CustomStep, XMLable):
         if not he:
             et = 0 * UR("seconds")
             eht = 0 * UR("seconds")
-        return Step(ht, ts, collect, et, etc, eht, ehtc, [], True)
+        return Step(ht, ts, collect, et, etc, 1, eht, ehtc, 1, [], True)
 
     def to_xml(self, **kwargs: Any) -> ET.Element:
         assert not kwargs
@@ -734,7 +792,7 @@ class Step(CustomStep, XMLable):
         return e
 
     @classmethod
-    def from_scpicommand(cls, sc: SCPICommand):
+    def from_scpicommand(cls, sc: SCPICommand) -> Step:
         coms = [x.specialize() for x in cast(Sequence[SCPICommand], sc.args[1])]
 
         # FIXME: When willing to require Python 3.10, this can be cleaned up with a match statement.
@@ -808,7 +866,11 @@ class Stage(XMLable, ProtoCommand):
     _classname: ClassVar[str] = "Stage"
     _names: ClassVar[Sequence[str]] = ("STAGe",)
 
-    def __eq__(self, other: object):
+    @property
+    def step(self) -> _NumOrRefIndexer[CustomStep]:
+        return _NumOrRefIndexer(list(self.steps))
+
+    def __eq__(self, other: object) -> bool:
         if not isinstance(other, Stage):
             return False
         if self.__class__ != other.__class__:
@@ -838,7 +900,7 @@ class Stage(XMLable, ProtoCommand):
         temperature_step: float | str | pint.Quantity[float] | None = None,
         collect: bool | None = None,
         filters: Sequence[str | FilterSet] = tuple(),
-    ):
+    ) -> Stage:
         """Hold at a series of temperatures, from one to another.
 
         Parameters
@@ -962,7 +1024,7 @@ class Stage(XMLable, ProtoCommand):
         step_time: int | str | pint.Quantity[int] | None = None,
         collect: bool | None = None,
         filters: Sequence[str | FilterSet] = tuple(),
-    ):
+    ) -> Stage:
         """Hold at a temperature for a set amount of time, with steps of a configurable fixed time.
 
 
@@ -1137,13 +1199,21 @@ class Stage(XMLable, ProtoCommand):
 
         return data
 
-    def to_scpicommand(self, stageindex=None, **kwargs) -> SCPICommand:
+    def to_scpicommand(
+        self, stageindex: int | str | None = None, **kwargs: Any
+    ) -> SCPICommand:
         opts = {}
-        args = []
+        args: list[int | str | list[SCPICommand]] = []
         if self.repeat != 1:
             opts["repeat"] = self.repeat
-        args.append(self.index or stageindex)
-        args.append(self.label or f"STAGE_{self.index or stageindex}")
+        if self.index:
+            index_to_use: int | str = self.index
+        elif stageindex:
+            index_to_use = stageindex
+        else:
+            raise ValueError("No index.")
+        args.append(index_to_use)
+        args.append(self.label or f"STAGE_{index_to_use}")
         args.append(
             [
                 step.to_scpicommand(stepindex=i + 1, **kwargs)
@@ -1154,7 +1224,7 @@ class Stage(XMLable, ProtoCommand):
         return SCPICommand("STAGe", *args, comment=None, **opts)
 
     @classmethod
-    def from_scpicommand(cls, sc: SCPICommand, **kwargs) -> Stage:
+    def from_scpicommand(cls, sc: SCPICommand, **kwargs: Any) -> Stage:
         c = cls(
             [
                 cast(CustomStep, x.specialize(**kwargs))
@@ -1209,7 +1279,7 @@ class Stage(XMLable, ProtoCommand):
 
         return e
 
-    def info_str(self, index) -> str:
+    def info_str(self, index: int | None = None) -> str:
         if self.repeat > 1:
             adds = "s"
         else:
@@ -1283,6 +1353,9 @@ class Protocol(ProtoCommand):
     ----------
     stages: Iterable[Stage]
         The stages of the protocol, likely :class:`Stage`.
+    stage: _NumOrRefIndexer[Stage]
+        A more convenient way of accessing the stages of the protocol, with
+        numbering that matches the machine.
     name: str | None
         A protocol name. If not set, a timestamp will be used, unlike AB's uuid.
     volume: float
@@ -1294,9 +1367,13 @@ class Protocol(ProtoCommand):
     filters: Sequence[str]
         A list of default filters that can be used by any collection commands
         that don't specify their own.
+    prerun: Sequence[SCPICommand]
+        Sets PRERUN.  *DO NOT USE THIS UNLESS YOU KNOW WHAT YOU ARE DOING*.
+    postrun: Sequence[SCPICommand]
+        Sets POSTRUN. *DO NOT USE THIS UNLESS YOU KNOW WHAT YOU ARE DOING*.
     """
 
-    stages: Iterable[Stage]
+    stages: list[Stage] = attr.field(factory=list)
     name: str = attr.field(factory=lambda: "Prot_" + _nowuuid())
     volume: float = 50.0
     runmode: str = "standard"  # standard or fast... need to deal with this
@@ -1306,9 +1383,22 @@ class Protocol(ProtoCommand):
         on_setattr=attr.setters.convert,
     )
     covertemperature: float = 105.0
+    prerun: Sequence[SCPICommandLike] = attr.field(factory=list)
+    postrun: Sequence[SCPICommandLike] = attr.field(factory=list)
     _classname: str = "Protocol"
     _names: ClassVar[Sequence[str]] = ("PROTocol", "PROT")
 
+    @property
+    def stage(self) -> _NumOrRefIndexer[Stage]:
+        """
+        A more convenient view of :any:`Protocol.stages`.  This allows one-indexed access,
+        such that `protocol.stage[5] == protocol.stages[6]` is stage 5 of the protocol, in
+        the interpretation of tha machine.  Indexing can use slices, and is *inclusive*, so
+        `protocol.stage[5:6]` returns stages 5 and 6.  Getting, setting, and appending
+        stages are all supported through this interface.
+        """
+        return _NumOrRefIndexer(self.stages)
+
     def to_scpicommand(self, **kwargs: Any) -> SCPICommand:
         assert not kwargs
         args: list[int | str | Sequence[SCPICommand]] = []
@@ -1318,26 +1408,57 @@ class Protocol(ProtoCommand):
         if self.runmode is not None:
             opts["runmode"] = self.runmode
         args.append(self.name)
-        args.append(
-            [
-                stage.to_scpicommand(
-                    filters=self.filters, stageindex=i + 1, default_filters=self.filters
-                )
-                for i, stage in enumerate(self.stages)
-            ]
-        )
+        stages: list[SCPICommand] = []
+        if self.prerun:
+            stages.append(
+                SCPICommand("PRERun", [s.to_scpicommand() for s in self.prerun])
+            )
+
+        stages += [
+            stage.to_scpicommand(
+                filters=self.filters, stageindex=i + 1, default_filters=self.filters
+            )
+            for i, stage in enumerate(self.stages)
+        ]
+
+        if self.postrun:
+            stages.append(
+                SCPICommand("POSTRun", [s.to_scpicommand() for s in self.postrun])
+            )
+
+        args.append(stages)
+
         return SCPICommand("PROTocol", *args, comment=None, **opts)
 
     @classmethod
     def from_scpicommand(cls: Type[Protocol], sc: SCPICommand) -> Protocol:
-        c = cls(
-            [
-                cast(Stage, x.specialize())
-                for x in cast(Sequence[SCPICommand], sc.args[1])
-            ],
-            name=cast(str, sc.args[0]),
-            **sc.opts,  # type: ignore
-        )  # type: ignore
+        stage_commands = [x for x in cast(Sequence[SCPICommand], sc.args[1])]
+
+        if stage_commands[0].command.upper() == "PRERUN":
+            prerun = cast(Sequence[SCPICommand], stage_commands[0].args[0])
+            assert isinstance(prerun, Sequence)
+            del stage_commands[0]
+        else:
+            prerun = []
+        if stage_commands[-1].command.upper() == "POSTRUN":
+            postrun = cast(Sequence[SCPICommand], stage_commands[-1].args[0])
+            assert isinstance(postrun, Sequence)
+            del stage_commands[-1]
+        else:
+            postrun = []
+
+        stages = [cast(Stage, x.specialize()) for x in stage_commands]
+        for s in stages:
+            if not isinstance(s, Stage):
+                raise ValueError
+
+        name = sc.args[0]
+        if not isinstance(name, str):
+            raise ValueError
+
+        opts = sc.opts
+
+        c = cls(stages, name, prerun=prerun, postrun=postrun, **opts)  # type: ignore
 
         dfilt: list[FilterSet] = []
 
@@ -1481,7 +1602,9 @@ class Protocol(ProtoCommand):
         stages = [Stage.from_xml(x) for x in e.findall("TCStage")]
         return Protocol(stages, protoname, svol, runmode, filters, covertemperature)
 
-    def to_xml(self, covertemperature=105.0) -> tuple[ET.ElementTree, ET.ElementTree]:
+    def to_xml(
+        self, covertemperature: float = 105.0
+    ) -> tuple[ET.ElementTree, ET.ElementTree]:
         te = ET.ElementTree(ET.Element("TCProtocol"))
         tqe = ET.ElementTree(ET.Element("QSTCProtocol"))
 
