@@ -5,18 +5,19 @@
 """SCPI Command class and parsing"""
 
 from __future__ import annotations
-from abc import ABC, abstractmethod
-import re
 
+import re
 import shlex
 import textwrap
+from abc import ABC, abstractmethod
 from dataclasses import dataclass
 from enum import Enum
 from typing import Any, Literal, Sequence, Type, TypeVar, cast
 
 import numpy as np
 import pyparsing as pp
-from pyparsing import ParserElement, pyparsing_common as ppc
+from pyparsing import ParserElement
+from pyparsing import pyparsing_common as ppc
 
 pp.ParserElement.setDefaultWhitespaceChars("")
 
@@ -100,6 +101,21 @@ _arglist.setParseAction(
     )
 )
 
+
+@dataclass
+class ArgList:
+    "A representation of an SCPI list of options (-key=value) and arguments."
+    opts: dict[str, bool | int | float | str]
+    args: list[bool | int | float | str]
+
+    @classmethod
+    def from_string(cls, argument_string: str) -> ArgList:
+        """Parse an SCPI argument string."""
+        return cast(ArgList, _arglist.parseString(argument_string)[0])
+
+
+_NULLIST = ArgList({}, [])
+
 _commentstring: pp.ParserElement = pp.Combine(
     pp.Regex(r"\s*#\s?").suppress()
     + pp.Regex(r"[^\n]+")
@@ -109,7 +125,7 @@ _commentstring: pp.ParserElement = pp.Combine(
 _command: pp.ParserElement = cast(
     pp.ParserElement,
     (
-        ppc.identifier("command")
+        pp.Combine(pp.Regex("[A-Za-z:_]+"))("command")
         + _ws_or_end
         + pp.Optional(_arglist("arglist"))
         + pp.Optional(_commentstring("comment"))
@@ -118,8 +134,8 @@ _command: pp.ParserElement = cast(
         lambda toks: SCPICommand(
             toks["command"],
             comment=toks.get("comment", None),
-            *toks["arglist"].args,
-            **toks["arglist"].opts,
+            *toks.get("arglist", _NULLIST).args,  # FIXME
+            **toks.get("arglist", _NULLIST).opts,
         )
     ),
 )
@@ -175,18 +191,6 @@ class AccessLevel(Enum):
         return self.value
 
 
-@dataclass
-class ArgList:
-    "A representation of an SCPI list of options (-key=value) and arguments."
-    opts: dict[str, bool | int | float | str]
-    args: list[bool | int | float | str]
-
-    @classmethod
-    def from_string(cls, argument_string: str) -> ArgList:
-        """Parse an SCPI argument string."""
-        return cast(ArgList, _arglist.parseString(argument_string)[0])
-
-
 T = TypeVar("T")
 
 
@@ -230,7 +234,7 @@ class SCPICommand(SCPICommandLike):
     ]
     comment: str | None
 
-    def __eq__(self, other) -> bool:
+    def __eq__(self, other: Any) -> bool:
         if not isinstance(other, SCPICommand):
             return False
 
