@@ -2,12 +2,18 @@
 # SPDX-License-Identifier: AGPL-3.0-only
 
 from pathlib import Path
-import numpy as np
 
+import numpy as np
 import pytest
 
 from qslib import Experiment
-from qslib.normalization import NormToMaxPerWell, NormToMeanPerWell
+from qslib.processors import (
+    NormToMaxPerWell,
+    NormToMeanPerWell,
+    SmoothEMWMean,
+    SmoothWindowMean,
+    SubtractByMeanPerWell,
+)
 
 
 @pytest.fixture(scope="module")
@@ -29,20 +35,42 @@ def exp_reloaded(
     return Experiment.from_file(tmp_path / "test_loaded.eds")
 
 
-def test_props(exp: Experiment, exp_reloaded: Experiment):
+def test_props(exp: Experiment, exp_reloaded: Experiment) -> None:
     assert exp.name == "2020-02-20_170706"
 
     assert exp.info() == str(exp) == exp.summary()
 
 
-def test_reload(exp: Experiment, exp_reloaded: Experiment):
+def test_reload(exp: Experiment, exp_reloaded: Experiment) -> None:
     assert (exp.welldata == exp_reloaded.welldata).all().all()
     assert exp.name == exp_reloaded.name
     assert exp.protocol == exp_reloaded.protocol
     assert exp.plate_setup == exp_reloaded.plate_setup
 
 
-def test_plots(exp: Experiment):
+def test_plot_ntmpw_smoothmw(exp: Experiment) -> None:
+    axf, axt = exp.plot_over_time(
+        process=[SmoothWindowMean(4), NormToMeanPerWell(1)], annotate_stage_lines=False
+    )
+    assert axf.get_ylabel() == "fluorescence (window mean 4, norm. to mean)"
+
+
+def test_plot_emw_maxperwell(exp: Experiment) -> None:
+    axf, axt = exp.plot_over_time(
+        process=[SmoothEMWMean(alpha=0.1), NormToMaxPerWell(1)],
+        annotate_stage_lines=False,
+    )
+    assert axf.get_ylabel() == "fluorescence (EMW-smoothed, norm. to max)"
+
+
+def test_plot_subtrbymean(exp: Experiment) -> None:
+    axf, axt = exp.plot_over_time(
+        process=SubtractByMeanPerWell(1), annotate_stage_lines=False
+    )
+    assert axf.get_ylabel() == "fluorescence (subtr. by mean)"
+
+
+def test_plots(exp: Experiment) -> None:
     axf, axt = exp.plot_over_time(
         legend=False, figure_kw={"constrained_layout": False}, annotate_stage_lines=True
     )
@@ -93,5 +121,5 @@ def test_plots(exp: Experiment):
     ax2 = exp.plot_protocol()
 
 
-def test_rawquant(exp: Experiment):
+def test_rawquant(exp: Experiment) -> None:
     exp.rawdata.loc[:, :]
